@@ -11,9 +11,14 @@ TAG_DEFINITIONS = {
     'strikethrough': {'strikethrough': True},
     'highlight': {'background': 'yellow'},
     'link': {'underline': Pango.Underline.SINGLE, 'foreground': 'blue'},
-    'header': {'scale': 1.5, 'weight': 500, 'pixels-above-lines': 14, 'pixels-below-lines': 4}
+    'header': {'scale': 1.5, 'weight': 500, 'pixels-above-lines': 14, 'pixels-below-lines': 4},
+    'small': {'scale': .8333333333333},
+    'large': {'scale': 1.2},
+    'larger': {'scale': 1.44}
 }
 
+FONT_SCALES = ['small', 'normal', 'large', 'larger']
+
 class GenericAction(object):
     def maybe_join(self, new_action):
         return False
@@ -310,7 +315,7 @@ class ShiftAction(GenericAction):
         else:
             self.shift_down()
 
-# Used to combine multiple actions into one single undable action. Actions should be passed in the same order in which
+# Used to combine multiple actions into one single undoable action. Actions should be passed in the same order in which
 # they were performed. Failure to do so could result in order getting mixed up in the buffer.
 class CompositeAction(GenericAction):
     def __init__(self, *args):
@@ -325,6 +330,53 @@ class CompositeAction(GenericAction):
         for action in self.child_actions:
             action.redo()
 
+# we need to subclass the check button to add a Gdk.Window so that we can change the cursor
+class CheckBox(Gtk.CheckButton):
+    def __init__(self, **kwargs):
+        super(CheckBox, self).__init__(**kwargs)
+
+        self.event_window = None
+
+        self.set_has_window(False)
+
+    def do_realize(self):
+        Gtk.CheckButton.do_realize(self)
+
+        parent_window = self.get_parent_window()
+        self.set_window(parent_window)
+
+        allocation = self.get_allocation()
+        attributes = Gdk.WindowAttr()
+        attributes.cursor = Gdk.Cursor.new_from_name(self.get_display(), 'default')
+        attributes.window_type = Gdk.WindowType.CHILD
+        attributes.x = allocation.x
+        attributes.y = allocation.y
+        attributes.width = allocation.width
+        attributes.height = allocation.height
+        attributes.wclass = Gdk.WindowWindowClass.INPUT_ONLY
+        attributes.event_mask = self.get_events() | Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.TOUCH_MASK | Gdk.EventMask.ENTER_NOTIFY_MASK | Gdk.EventMask.LEAVE_NOTIFY_MASK | Gdk.EventMask.POINTER_MOTION_MASK
+
+        attributes_mask = Gdk.WindowAttributesType.X | Gdk.WindowAttributesType.Y | Gdk.WindowAttributesType.CURSOR
+
+        self.event_window = Gdk.Window.new(parent_window, attributes, attributes_mask)
+        self.register_window(self.event_window)
+
+    def do_map(self):
+        Gtk.CheckButton.do_map(self)
+
+        self.event_window.show()
+
+    def do_unmap(self):
+        Gtk.CheckButton.do_unmap(self)
+
+        self.event_window.hide()
+
+    def do_size_allocate(self, allocation):
+        Gtk.CheckButton.do_size_allocate(self, allocation)
+
+        if self.event_window:
+            self.event_window.move(allocation.x, allocation.y)
+
 class NoteBuffer(Gtk.TextBuffer):
     # These values should not be modified directly.
     # internal_action_count keeps the "content-changed" signal from firing while the buffer performs several actions. It
@@ -363,8 +415,10 @@ class NoteBuffer(Gtk.TextBuffer):
     def __init__(self):
         super(NoteBuffer, self).__init__()
 
+        self.tags = []
+
         for name, attributes in TAG_DEFINITIONS.items():
-            self.create_tag(name, **attributes)
+            self.tags.append(self.create_tag(name, **attributes))
 
         self.connect('delete-range', self.on_delete)
         self.connect('begin-user-action', self.begin_composite_action)
@@ -421,44 +475,25 @@ class NoteBuffer(Gtk.TextBuffer):
         return InternalActionHandler()
 
     def get_internal_markup(self):
-        current_tags = []
+        on_tags = []
+        off_tags = self.tags.copy()
         text = ''
 
         current_iter = self.get_iter_at_offset(0)
         while True:
-            # if not all tags are closed, we still need to keep track of them, but leaving them in the list will
-            # cause an infinite loop, so we hold on to them in unclosed_tags and re-add them after exiting the loop
-            unclosed_tags = []
+            # first we close any open tags that don't continue on to the current character
+            for tag in on_tags:
+                if not current_iter.has_tag(tag):
+                    text += '#tag:%s:' % tag.props.name
+                    off_tags.append(tag)
+                    on_tags.remove(tag)
 
-            # end tags
-            tags = current_iter.get_toggled_tags(False)
-            while len(current_tags) and len(tags):
-                tag = current_tags.pop()
-                name = tag.props.name
-
-                if not name or name not in TAG_DEFINITIONS:
-                    continue
-
-                if len(tags) == 0 or tag not in tags:
-                    unclosed_tags.append(tag)
-                    continue
-
-                text += '#tag:%s:' % name
-                tags.remove(tag)
-
-            current_tags += unclosed_tags
-
-            # start tags
-            tags = current_iter.get_toggled_tags(True)
-            while len(tags):
-                tag = tags.pop()
-                name = tag.props.name
-
-                if not name or name not in TAG_DEFINITIONS:
-                    continue
-
-                text += '#tag:%s:' % tag.props.name
-                current_tags.append(tag)
+            # next we open any tags that start with the current character
+            for tag in off_tags:
+                if current_iter.has_tag(tag):
+                    text += '#tag:%s:' % tag.props.name
+                    on_tags.append(tag)
+                    off_tags.remove(tag)
 
             current_char = current_iter.get_char()
             if current_char == '#':
@@ -478,9 +513,8 @@ class NoteBuffer(Gtk.TextBuffer):
             if not current_iter.forward_char():
                 break
 
-        # If the tag goes to the end of the text, the 'toggle-off' point will technically be after the end of the text,
-        # so we wont pick it up above. Instead, we'll just close all tags and assume they're supposed to go to the end.
-        for tag in current_tags:
+        # If there are any open tags at this point, it means they go to the end of the text, so close them
+        for tag in on_tags:
             text += '#tag:%s:' % tag.props.name
 
         return text
@@ -597,40 +631,57 @@ class NoteBuffer(Gtk.TextBuffer):
 
                     action = CompositeAction(action, obj_action)
 
-                    location.assign(self.get_iter_at_offset(position-1))
+            # location gets shifted by one when Gtk.TextBuffer.do_insert_text() gets called, and becomes invalidated by
+            # adding a check/bullet, so we need to put it back to where it was before
+            location.assign(self.get_iter_at_offset(position))
 
-            start = location.copy()
-            start.backward_char()
+            next_char = location.copy()
+            next_char.forward_char()
 
             # check tag toggles
             if length == 1:
+                font_scale_changed = False
                 for tag_name in self.tag_toggles:
-                    if location.has_tag(self.get_tag_table().lookup(tag_name)):
-                        action = CompositeAction(action, self.strip_tag(tag_name, start, location))
-                    else:
-                        action = CompositeAction(action, self.add_tag(tag_name, start, location))
+                    # if the tag is a font scale, remove all other font scales
+                    if tag_name in FONT_SCALES:
+                        font_scale_changed = True
+                        for scale_tag_name in FONT_SCALES:
+                            if scale_tag_name != 'normal':
+                                action = CompositeAction(action, self.strip_tag(scale_tag_name, location, next_char))
 
-                for tag in start.get_toggled_tags(False):
+                    prev = location.copy()
+                    if tag_name == 'normal':
+                        # the 'normal' tag doesn't really exist but we still use it for removing other font scale tags
+                        pass
+                    elif prev.backward_char() & prev.has_tag(self.get_tag_table().lookup(tag_name)):
+                        action = CompositeAction(action, self.strip_tag(tag_name, location, next_char))
+                    else:
+                        action = CompositeAction(action, self.add_tag(tag_name, location, next_char))
+
+                for tag in location.get_toggled_tags(False):
                     tag_name = tag.props.name
                     if tag_name is None:
                         continue
 
+                    # don't continue header tag onto new line
                     if text == '\n' and tag_name == 'header':
                         continue
 
-                    if tag_name not in self.tag_toggles:
-                        action = CompositeAction(action, self.add_tag(tag_name, start, location))
+                    # don't continue the tag if it was toggled off
+                    # also don't continue if it's a font scale tag and another was just applied
+                    if tag_name not in self.tag_toggles and (tag_name not in FONT_SCALES or not font_scale_changed):
+                        action = CompositeAction(action, self.add_tag(tag_name, location, next_char))
+
+                self.tag_toggles = []
 
             if not self.props.can_undo or not self.undo_actions[-1].maybe_join(action):
                 self.add_undo_action(action)
 
         if text in ['\n', '\t', ' ', '.', ',', ';', ':']:
-            end = location.copy()
-            end.backward_char()
-            pre_text = self.get_slice(self.get_start_iter(), end, True)
+            pre_text = self.get_slice(self.get_start_iter(), location, True)
             match = get_url_start(pre_text)
             if match:
-                self.add_undo_action(self.add_tag('link', self.get_iter_at_offset(match.start()), end))
+                self.add_undo_action(self.add_tag('link', self.get_iter_at_offset(match.start()), location))
 
     def on_delete(self, buffer, start, end):
         if self.internal_action_count:
@@ -709,31 +760,55 @@ class NoteBuffer(Gtk.TextBuffer):
             (start, end) = self.get_selection_bounds()
 
             current_iter = start.copy()
-            all_has_tag = True
-            while current_iter.compare(end) < 0:
-                if not current_iter.has_tag(self.get_tag_table().lookup(tag_name)):
-                    all_has_tag = False
-                    break
+            remove_tag = True
+            if tag_name in FONT_SCALES:
+                remove_tag = False
+            else:
+                tag = self.get_tag_table().lookup(tag_name)
+                while current_iter.compare(end) < 0:
+                    if not current_iter.has_tag(tag):
+                        remove_tag = False
+                        break
 
-                current_iter.forward_char()
+                    current_iter.forward_char()
 
-            # remove the tag if the whole selection already has it
-            if all_has_tag:
+            # remove the tag if the whole selection already has it (unless it's a font scale tag)
+            if remove_tag:
                 self.add_undo_action(self.strip_tag(tag_name, start, end))
             else:
                 self.add_undo_action(self.add_tag(tag_name, start, end))
         else:
+            # if it's a font scale, we want to remove any existing font scale first
+            # this has the side effect that font scales don't toggle like everything else, which is exactly what we want
+            if tag_name in FONT_SCALES:
+                for name in self.tag_toggles:
+                    if name in FONT_SCALES:
+                        self.tag_toggles.remove(name)
+
             if tag_name in self.tag_toggles:
                 self.tag_toggles.remove(tag_name)
             else:
                 self.tag_toggles.append(tag_name)
 
     def add_tag(self, tag_name, start, end):
-        action = TagAction(self, tag_name, start, end)
-        self.apply_tag_by_name(tag_name, start, end)
-        self.trigger_changed()
+        actions = []
+        if tag_name in FONT_SCALES:
+            for font_tag_name in FONT_SCALES:
+                if font_tag_name == tag_name or font_tag_name == 'normal':
+                    continue
 
-        return action
+                action = self.strip_tag(font_tag_name, start, end)
+                if len(action.ranges) > 0:
+                    actions.append(action)
+
+        if tag_name != 'normal':
+            actions.append(TagAction(self, tag_name, start, end))
+            self.apply_tag_by_name(tag_name, start, end)
+        self.trigger_changed()
+        if len(actions) == 1:
+            return actions[0]
+        else:
+            return CompositeAction(*actions)
 
     def strip_tag(self, tag_name, start, end):
         action = TagAction(self, tag_name, start, end, False)
@@ -745,7 +820,7 @@ class NoteBuffer(Gtk.TextBuffer):
     def add_check_button(self, a_iter, checked=False):
         with self.internal_action():
             anchor = self.create_child_anchor(a_iter)
-            check_button = Gtk.CheckButton(visible=True, active=checked, margin_right=5, margin_top=5)
+            check_button = CheckBox(visible=True, active=checked, margin_right=5, margin_top=5)
             check_button.connect('toggled', self.trigger_changed)
             self.view.add_child_at_anchor(check_button, anchor)
 
