@@ -2,7 +2,6 @@
 
 import json
 import os
-import random
 import sys
 
 import gi
@@ -26,6 +25,8 @@ APPLICATION_ID = 'org.x.sticky'
 STYLE_SHEET_PATH = '/usr/share/sticky/sticky.css'
 SCHEMA = 'org.x.sticky'
 
+UPDATE_DELAY = 1
+
 FONT_SCALES = [
     ('small', _("Small Text"), 'small'),
     ('normal', _("Normal Text"), 'medium'),
@@ -91,13 +92,15 @@ class Note(Gtk.Window):
 
         self.showing = False
         self.is_pinned = False
+        self.changed_timer_id = 0
+        self.invalid_cache = False
 
         self.x = info.get('x', 0)
         self.y = info.get('y', 0)
         self.height = info.get('height', self.app.settings.get_uint('default-height'))
         self.width = info.get('width', self.app.settings.get_uint('default-width'))
         title = info.get('title', '')
-        text = info.get('text', '')
+        self.cached_text = info.get('text', '')
         self.color = info.get('color', self.app.settings.get_string('default-color'))
 
         super(Note, self).__init__(
@@ -111,8 +114,19 @@ class Note(Gtk.Window):
             name='sticky-note'
         )
 
-        if self.color == 'random':
-            self.color = random.choice(list(COLORS.keys()))
+        if self.color == 'cycle':
+            if self.app.settings.get_string('last-color') == '':
+                last_color = self.color
+            else:
+                last_color = self.app.settings.get_string('last-color')
+
+            color_keys = list(COLORS)
+            try:
+                self.color = color_keys[color_keys.index(last_color) + 1]
+            except (ValueError, IndexError):
+                self.color = color_keys[0]
+
+            self.app.settings.set_string('last-color', self.color)
 
         context = self.get_style_context()
         context.add_class(self.color)
@@ -157,7 +171,7 @@ class Note(Gtk.Window):
 
         add_icon = Gtk.Image.new_from_icon_name('sticky-add', Gtk.IconSize.BUTTON)
         add_button = Gtk.Button(image=add_icon, relief=Gtk.ReliefStyle.NONE, name='window-button', valign=Gtk.Align.CENTER)
-        add_button.connect('clicked', self.app.new_note)
+        add_button.connect('clicked', self.app.new_note, self)
         add_button.connect('button-press-event', self.on_title_click)
         add_button.set_tooltip_text(_("New Note"))
         self.title_bar.pack_end(add_button, False, False, 0)
@@ -191,15 +205,15 @@ class Note(Gtk.Window):
         self.add(scroll)
         scroll.add(self.view)
 
-        self.buffer.set_from_internal_markup(text)
-        self.changed_id = self.buffer.connect('content-changed', self.changed)
+        self.buffer.set_from_internal_markup(self.cached_text)
+        self.changed_id = self.buffer.connect('content-changed', self.queue_update, True)
 
         self.app.settings.connect('changed::font', self.set_font)
         self.set_font()
 
         self.create_format_menu(color_button, text_button)
 
-        self.connect('configure-event', self.handle_update)
+        self.connect('configure-event', self.on_size_position_changed)
         self.connect('show', self.on_show)
         self.connect('window-state-event', self.update_window_state)
 
@@ -210,7 +224,7 @@ class Note(Gtk.Window):
     def test(self, *args):
         self.buffer.test()
 
-    def handle_update(self, *args):
+    def on_size_position_changed(self, *args):
         if self.showing:
             self.showing = False
             return
@@ -224,7 +238,8 @@ class Note(Gtk.Window):
         self.y = new_y
         self.height = new_height
         self.width = new_width
-        self.emit('update')
+
+        self.queue_update()
 
     def on_show(self, *args):
         self.showing = True
@@ -328,20 +343,33 @@ class Note(Gtk.Window):
         self.present_with_time(time)
         self.move(self.x, self.y)
 
-    def changed(self, *args):
+    def queue_update(self, b=None, invalidate_cache=False):
+        self.invalid_cache = invalidate_cache
+
+        if self.changed_timer_id:
+            GLib.source_remove(self.changed_timer_id)
+
+        self.changed_timer_id = GLib.timeout_add_seconds(UPDATE_DELAY, self.trigger_update)
+
+    def trigger_update(self):
+        self.changed_timer_id = 0
+
         self.emit('update')
 
     def get_info(self):
-        (x, y) = self.get_position()
+        if self.invalid_cache:
+            self.cached_text = self.buffer.get_internal_markup()
+            self.invalid_cache = False
+
         (width, height) = self.get_size()
         info = {
-            'x': x,
-            'y': y,
-            'height': height,
-            'width': width,
+            'x': self.x,
+            'y': self.y,
+            'height': self.height,
+            'width': self.width,
             'color': self.color,
             'title': self.title.get_text(),
-            'text': self.buffer.get_internal_markup()
+            'text': self.cached_text
         }
 
         return info
@@ -365,6 +393,10 @@ class Note(Gtk.Window):
         edit_title.connect('activate', self.set_title)
         popup.append(edit_title)
 
+        duplicate_item = Gtk.MenuItem(label=_("Duplicate Note"), visible=True)
+        duplicate_item.connect('activate', self.duplicate)
+        popup.append(duplicate_item)
+
         remove_item = Gtk.MenuItem(label=_("Delete Note"), visible=True)
         remove_item.connect('activate', self.remove)
         popup.append(remove_item)
@@ -483,11 +515,15 @@ class Note(Gtk.Window):
     def remove(self, *args):
         # this is ugly but I'm not sure how to make it look better :)
         if (self.app.settings.get_boolean('disable-delete-confirm') or
+            (not self.title.get_text() and self.buffer.get_char_count() == 0) or
             confirm(_("Delete Note"), _("Are you sure you want to remove this note?"),
                     self, self.app.settings, 'disable-delete-confirm')):
             self.emit('removed')
             self.destroy()
 
+    def duplicate(self, *args):
+        self.app.duplicate_note(self)
+
     def set_title(self, *args):
         self.title_text = self.title.get_text()
         self.title_box.remove(self.title)
@@ -553,12 +589,12 @@ class SettingsWindow(XApp.PreferencesWindow):
         try:
             colors = [(x, y) for x, y in COLORS.items()]
             colors.append(('sep', ''))
-            colors.append(('random', _('Random')))
+            colors.append(('cycle', _('Cycle Colors')))
 
             page.pack_start(GSettingsComboBox(_("Default color"), SCHEMA, 'default-color', options=colors, valtype=str, separator='sep'), False, False, 0)
         except Exception as e:
             colors = [(x, y) for x, y in COLORS.items()]
-            colors.append(('random', _('Random')))
+            colors.append(('cycle', _('Cycle Colors')))
 
             page.pack_start(GSettingsComboBox(_("Default color"), SCHEMA, 'default-color', options=colors, valtype=str), False, False, 0)
 
@@ -644,6 +680,11 @@ class Application(Gtk.Application):
         if self.settings.get_boolean('first-run'):
             self.first_run()
 
+        # Backwards compatibility
+        # - Update random color option to cycle
+        if self.settings.get_string('default-color') == 'random':
+            self.settings.set_string('default-color', 'cycle')
+
         self.file_handler.connect('lists-changed', self.on_lists_changed)
         self.group_update_id = self.file_handler.connect('group-changed', self.on_group_changed)
         self.file_handler.connect('group-name-changed', self.on_group_name_changed)
@@ -768,6 +809,12 @@ class Application(Gtk.Application):
 
             menu.append(Gtk.SeparatorMenuItem())
 
+            item = Gtk.MenuItem(label=_("Preferences"))
+            item.connect('activate', self.open_settings_window)
+            menu.append(item)
+
+            menu.append(Gtk.SeparatorMenuItem())
+
             item = Gtk.MenuItem(label=_("Quit"))
             item.connect('activate', self.quit_app)
             menu.append(item)
@@ -814,19 +861,26 @@ class Application(Gtk.Application):
         for note in self.notes:
             note.restore(time)
 
+        if len(self.notes) == 0:
+            self.new_note()
+
         self.notes_hidden = False
         self.update_dummy_window()
 
     def hide_notes(self):
         for note in self.notes:
             note.hide()
-
+        
         self.notes_hidden = True
         self.update_dummy_window()
 
-    def new_note(self, *args):
-        x = 40
-        y = 40
+    def new_note(self, button, parent=None):
+        if parent is None:
+            x = 40
+            y = 40
+        else:
+            x = parent.x + 20
+            y = parent.y + 20
         while(True):
             found = False
             for note_info in self.file_handler.get_note_list(self.note_group):
@@ -838,6 +892,10 @@ class Application(Gtk.Application):
             x += 20
             y += 20
         info = {'x': x, 'y': y}
+
+        self.add_note(info)
+
+    def add_note(self, info):
         note = self.generate_note(info)
         note.present_with_time(Gtk.get_current_event_time())
 
@@ -847,7 +905,7 @@ class Application(Gtk.Application):
             note.realize()
         note.get_window().raise_()
 
-        note.changed()
+        self.on_update()
 
     def generate_note(self, info={}):
         note = Note(self, self.dummy_window, info)
@@ -867,6 +925,13 @@ class Application(Gtk.Application):
         for note_info in self.file_handler.get_note_list(self.note_group):
             self.generate_note(note_info)
 
+    def duplicate_note(self, new_note):
+        new_note_info = new_note.get_info()
+        new_note_info['x'] += 50
+        new_note_info['y'] += 50
+
+        self.add_note(new_note_info)
+
     def focus_note(self, note_info):
         for note in self.notes:
             if note.get_info() == note_info:
@@ -1016,4 +1081,3 @@ if __name__ == "__main__":
     sticky = Application()
     sticky.autostart_mode = autostart_mode
     sticky.run()
-
